/*
 * Copyright 2009-2013 Aarhus University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.brics.tajs;

import dk.brics.tajs.analysis.Analysis;
import dk.brics.tajs.analysis.Context;
import dk.brics.tajs.analysis.State;
import dk.brics.tajs.flowgraph.FlowGraph;
import dk.brics.tajs.htmlparser.HTMLParser;
import dk.brics.tajs.htmlparser.JavaScriptSource;
import dk.brics.tajs.htmlparser.JavaScriptSource.EventHandlerJavaScriptSource;
import dk.brics.tajs.js2flowgraph.RhinoASTToFlowgraph;
import dk.brics.tajs.lattice.*;
import dk.brics.tajs.monitoring.IAnalysisMonitoring;
import dk.brics.tajs.options.ExperimentalOptions;
import dk.brics.tajs.options.Options;
import dk.brics.tajs.solver.CallGraph;
import dk.brics.tajs.solver.SolverSynchronizer;
import dk.brics.tajs.util.AnalysisException;
import dk.brics.tajs.util.Strings;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.jdom.Document;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import static dk.brics.tajs.util.Collections.newList;

/**
 * Main class for the TAJS program analysis.
 */
public class Main {
    public static enum Phase {
        LOADING_FILES,
        DATAFLOW_ANALYSIS,
        MESSAGES,
        STATISTICS,
        COVERAGE_SUMMARY,
        REACHABLE_VARIABLES_SUMMARY,
        CALL_GRAPH;
    }

    private static Logger logger = Logger.getLogger(Main.class);

	private Main() {}

	/**
	 * Runs the analysis on the given source files. Run without arguments to see
	 * the usage. Terminates with System.exit.
	 */
	public static void main(String[] args) {
		try {
			initLogging();
			Analysis a = init(args, null);
			if (a == null)
				System.exit(-1);
			run(a);
			System.exit(0);
		} catch (AnalysisException e) {
			e.printStackTrace();
			System.exit(-2);
		}
	}
	
	/**
	 * Resets all internal counters and caches.
	 */
	public static void reset() {
        ExperimentalOptions.ExperimentalOptionsManager.set(new ExperimentalOptions());
        Options.reset();
		BlockState.reset();
		Value.reset();
		Obj.reset();
		Strings.reset();
		ScopeChain.reset();
	}

	/**
	 * Reads the input and prepares an analysis object.
	 * @return analysis object, null if invalid input
	 * @throws AnalysisException if internal error
	 */
	public static Analysis init(String[] args, SolverSynchronizer sync) throws AnalysisException {
        Analysis analysis = new Analysis(sync);
        IAnalysisMonitoring monitoring = analysis.getMonitoring();

        boolean show_usage = false;
		Options.parse(args);
		List<String> files = Options.getArguments();
		
		if (files.isEmpty()) {
			System.out.println("No source files");
			show_usage = true;
		}
		if (show_usage) {
			System.out.println("TAJS - Type Analyzer for JavaScript");
			System.out.println("Copyright 2009-2013 Aarhus University\n");
			System.out.println("Usage: java -jar tajs-all.jar [OPTION]... [FILE]...\n");
			Options.describe();
			return null;
		}
		if(Options.isDebugEnabled())
			Options.dump();

		enterPhase(Phase.LOADING_FILES, monitoring);

		Document document = null;
		FlowGraph fg;
		try {
			List<String> html_files = newList();
			List<String> js_files = newList();
			for (String fn : files) {
				String l = fn.toLowerCase();
				if (l.endsWith(".html") || l.endsWith(".xhtml") || l.endsWith(".htm"))
					html_files.add(fn);
				else
					js_files.add(fn);
			}
			RhinoASTToFlowgraph builder = new RhinoASTToFlowgraph();
			if (!html_files.isEmpty()) {
				if (html_files.size() > 1)
					throw new AnalysisException("Only one HTML file can be analyzed at a time.");
				Options.enableIncludeDom(); // enable DOM if any HTML files are involved
				HTMLParser p = new HTMLParser();
				document = p.build(html_files.get(0));
                List<EventHandlerJavaScriptSource> eventList = p.getEventHandlerAttributeList();
                List<JavaScriptSource> scriptList = p.getScriptList();
//                List<HtmlSource> htmlSourceList = p.getHtmlSourceList(); // TODO: getHtmlSourceList not used (intended for DSL mode?)
                builder.buildFromHTMLFile(scriptList, eventList); // TODO: correct order of combination of the code/page fragments?
			} // TODO: support analyzing one .html plus a number of .js files as one program?
			if (!js_files.isEmpty()) {
				builder.buildFromJavaScriptFiles(js_files);
			}
			fg = builder.close();
            fg.check();
		} catch (IOException e) {
			logger.error("Unable to parse " + e.getMessage());
			return null;
		}

        if (sync != null)
            sync.setFlowGraph(fg);
        monitoring.registerFlowgraph(fg);
        if (Options.isFlowGraphEnabled())
            dumpFlowGraph(fg, false);

		analysis.getSolver().init(fg, document);

		leavePhase(Phase.LOADING_FILES, monitoring);
		return analysis;
	}

	/**
	 * Configures log4j.
	 */
	public static void initLogging() {
        Properties prop = new Properties();
        prop.put("log4j.rootLogger", "INFO, tajs"); // DEBUG / INFO / WARN / ERROR 
        prop.put("log4j.appender.tajs", "org.apache.log4j.ConsoleAppender");
        prop.put("log4j.appender.tajs.layout", "org.apache.log4j.PatternLayout");
        prop.put("log4j.appender.tajs.layout.ConversionPattern", "[%p %C{1}] %m%n");
        PropertyConfigurator.configure(prop);
	}

	/**
	 * Runs the analysis.
	 * @throws AnalysisException if internal error
	 */
	public static void run(Analysis analysis) throws AnalysisException {
        IAnalysisMonitoring monitoring = analysis.getMonitoring();
		long time = System.currentTimeMillis();
        long elapsed = 0;

        enterPhase(Phase.DATAFLOW_ANALYSIS, monitoring);
		analysis.getSolver().solve();
    	elapsed = (System.currentTimeMillis() - time);
		if (Options.isTimingEnabled())
        	logger.info("Analysis finished in " + elapsed + "ms");
		leavePhase(Phase.DATAFLOW_ANALYSIS, monitoring);

        // FIXME: does it make sense to have more than one state as the final state?
        Map<Context,State> states = analysis.getSolver().getAnalysisLatticeElement().getStates(analysis.getSolver().getFlowGraph().getMain().getOrdinaryExit());
        final State finalState;
        if(!states.isEmpty()){
            finalState = states.values().iterator().next();
            analysis.getMonitoring().setFinalAnalysisState(finalState);
        }
        CallGraph<State, Context, CallEdge<State>> callGraph = analysis.getSolver().getAnalysisLatticeElement().getCallGraph();
        analysis.getMonitoring().setFinalCallGraph(callGraph);

        if (Options.isFlowGraphEnabled())
            dumpFlowGraph(analysis.getSolver().getFlowGraph(), true);

        if (!Options.isNoMessages()) {
        	enterPhase(Phase.MESSAGES, monitoring);
        	analysis.getSolver().scan();
        	leavePhase(Phase.MESSAGES, monitoring);
        }

		if (Options.isStatisticsEnabled()) {
			enterPhase(Phase.STATISTICS, monitoring);
			logger.info(monitoring);
            System.out.println(callGraph.getCallGraphStatistics());
//			System.out.println(analysis.getOptimizer());
			System.out.println("BlockState: created=" + BlockState.getNumberOfStatesCreated() + ", makeWritableStore=" + BlockState.getNumberOfMakeWritableStoreCalls());
			System.out.println("Obj: created=" + Obj.getNumberOfObjsCreated() + ", makeWritableProperties=" + Obj.getNumberOfMakeWritablePropertiesCalls());
			System.out.println("Value cache: hits=" + Value.getNumberOfValueCacheHits() + ", misses=" + Value.getNumberOfValueCacheMisses() + ", finalSize=" + Value.getValueCacheSize());
			System.out.println("Value object set cache: hits=" + Value.getNumberOfObjectSetCacheHits() + ", misses=" + Value.getNumberOfObjectSetCacheMisses() + ", finalSize=" + Value.getObjectSetCacheSize());
			System.out.println("ScopeChain cache: hits=" + ScopeChain.getNumberOfCacheHits() + ", misses=" + ScopeChain.getNumberOfCacheMisses() + ", finalSize=" + ScopeChain.getCacheSize());
			System.out.println("Basic blocks: " + analysis.getSolver().getFlowGraph().getNumberOfBlocks());
			leavePhase(Phase.STATISTICS, monitoring);
		}

		if (Options.isCoverageEnabled()) {
			enterPhase(Phase.COVERAGE_SUMMARY, monitoring);
			monitoring.generateUnreachableMap();
			leavePhase(Phase.COVERAGE_SUMMARY, monitoring);
		}

		if (Options.isCallGraphEnabled()) {
			enterPhase(Phase.CALL_GRAPH, monitoring);
			logger.info(callGraph);
			File outdir = new File("out");
			if (!outdir.exists()) {
				outdir.mkdir();
			}
			String filename = "out" + File.separator + "callgraph.dot";
			try (FileWriter f = new FileWriter(filename)) {
				logger.info("Writing call graph to " + filename);
				callGraph.toDot(new PrintWriter(f));
			} catch (IOException e) {
				logger.error("Unable to write " + filename + ": " + e.getMessage());
			}
			leavePhase(Phase.CALL_GRAPH, monitoring);
		}
	}

	/**
	 * Outputs the flowgraph (in graphviz dot files).
	 */
	private static void dumpFlowGraph(FlowGraph g, boolean end) {
		try {
			File outdir = new File("out");
			if (!outdir.exists()) {
				outdir.mkdir();
			}
			String path = "out" + File.separator + "flowgraphs";
			File outdir2 = new File(path);
			if (!outdir2.exists()) {
				outdir2.mkdir();
			}
			try (PrintWriter pw = new PrintWriter(new FileWriter(path + File.separator + (end ? "final" : "initial") + ".dot"))) { 
				g.toDot(pw);
			} catch (IOException e) {
				throw new AnalysisException(e);
			}
			System.out.println(g.toString());
			File dir = new File(path);
			if (!dir.exists()) {
				dir.mkdir();
			}
			g.toDot(path, end);
		} catch (IOException e) {
			throw new AnalysisException(e);
		}
	}

	private static void enterPhase(Phase phase, IAnalysisMonitoring monitoring) {
        monitoring.beginPhase(phase);
		if (!Options.isQuietEnabled())
			logger.info("===========  " + prettyPhaseName(phase) + " ===========");
	}

    private static String prettyPhaseName(Phase phase) {
        switch (phase) {
            case LOADING_FILES:
                return "Loading files";
            case DATAFLOW_ANALYSIS:
                return "Data flow analysis";
            case MESSAGES:
                return "Messages";
            case STATISTICS:
                return "Statistics";
            case REACHABLE_VARIABLES_SUMMARY:
                return "Reachable variables summary";
            case COVERAGE_SUMMARY:
                return "Coverage summary";
            case CALL_GRAPH:
                return "Call graph";
            default:
                throw new RuntimeException("Unhandled phase enum: " + phase);
        }
    }

    @SuppressWarnings("unused")
	private static void leavePhase(Phase phase, IAnalysisMonitoring monitoring) {
        monitoring.endPhase(phase);
		// if (!Options.isTestEnabled())
		//   logger.info(">END:" + name);
	}
}
