package dk.brics.tajs;

import dk.brics.tajs.analysis.Analysis;
import dk.brics.tajs.csi.*;
import dk.brics.tajs.options.ExperimentalOptions;
import dk.brics.tajs.options.OptionValues;
import dk.brics.tajs.options.Options;
import dk.brics.tajs.util.AnalysisException;
import dk.brics.tajs.util.Pair;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;

public class MainJQueryDeterminacy {
    private enum Benchmark {
        A, B, C, D
    }

    private enum Mode {
        ALL, NO_BRANCH_PRUNING, NO_PARAMETER_SENSITIVITY, NO_LOOP_SPECIALIZATION, NO_FOR_IN_SPECIALIZATION, NO_CONTEXT_SENSITIVE_HEAP, NONE, NO_IMPROVED_STANDARD_LIBRARY
    }

    public static void main (String[] args) throws IOException {
        Main.initLogging();
        args = new String[]{"A", "ALL", "/home/esbena/eclipse_workspace/TAJS/test/jQueryCustomizedLoading/jquery-1.8.js-customized-minimally.js-sliced.js"};
        // ./$0 MODE SOURCE
        if (args.length == 3) {
            final Mode mode;
            final Path source;
            final Benchmark benchmark;
            try {
                benchmark = Benchmark.valueOf(args[0]);
                mode = Mode.valueOf(args[1]);
                source = Paths.get(args[2]);
            } catch (Exception e) {
                printUsage();
                System.exit(1);
                return;
            }
            if (!Files.exists(source)) {
                throw new IOException("No such source file: " + source);
            }
            Pair<OptionValues, ExperimentalOptions> options = makeOptions(benchmark, mode, source);

            // Prepare state for experiment
            Options.set(options.getFirst());
            ExperimentalOptions.ExperimentalOptionsManager.set(options.getSecond());

            SyntaxBasedLoopVariableIdentifierManager.ReadInAConditionAndDPAUseInBody identifier = new SyntaxBasedLoopVariableIdentifierManager.ReadInAConditionAndDPAUseInBody();
            SyntaxBasedLoopVariableIdentifierManager.init(identifier);
            ContextSensitivityInferenceManager.init(new LoopVariableSensitive(identifier), new PreciseArgumentsContextCreator());
            LiteralContextSensitivityManager.init(new ParameterDependencyStrategy());

            Analysis a = Main.init(new String[]{source.toString()}, null);
            if (a == null)
                throw new AnalysisException("Error during initialization");
            Main.run(a);
        } else {
            printUsage();
            System.exit(1);
            return;
        }
    }


    private static Pair<OptionValues, ExperimentalOptions> makeOptions (Benchmark benchmark, Mode mode, Path source) {
        Pair<OptionValues, ExperimentalOptions> baseOptions = makeBaseOptions();
        switch (mode) {
            case ALL:
                break;
            case NO_BRANCH_PRUNING:
                baseOptions.getFirst().disableControlSensitivity();
                break;
            case NO_PARAMETER_SENSITIVITY:
                baseOptions.getFirst().disableParameterSensitivity();
                break;
            case NO_LOOP_SPECIALIZATION:
                baseOptions.getFirst().disableNumericVariableSensitivity();
                break;
            case NO_FOR_IN_SPECIALIZATION:
                baseOptions.getFirst().disableForInSpecialization();
                break;
            case NO_CONTEXT_SENSITIVE_HEAP:
                baseOptions.getFirst().disableContextSensitiveHeap();
                break;
            case NO_IMPROVED_STANDARD_LIBRARY:
                baseOptions.getSecond().disable(ExperimentalOptions.Option.CONCRETE_SEMANTICS);
                break;
            case NONE:
                baseOptions = Pair.make(new OptionValues(), new ExperimentalOptions());
                baseOptions.getFirst().disableForInSpecialization();
                baseOptions.getFirst().disableControlSensitivity();
                baseOptions.getFirst().enableQuiet();
                baseOptions.getFirst().enableIncludeDom();
                baseOptions.getSecond().disable(ExperimentalOptions.Option.CONCRETE_SEMANTICS);
                break;
        }
        switch (benchmark) {
            case A:
                break;
            case B:
                break;
            case C:
                baseOptions.getFirst().enableUnreachable();
                break;
            case D:
                if (source.getFileName().toString().endsWith(".js")) {
                    baseOptions.getFirst().disableIncludeDom();
                }
                break;
        }
        return baseOptions;
    }

    private static Pair<OptionValues, ExperimentalOptions> makeBaseOptions () {
        final ExperimentalOptions experimentalOptions = new ExperimentalOptions(
                ExperimentalOptions.Option.MANUAL_ASSUMES, // enables some of the falsy/truthy control sensitivity
                ExperimentalOptions.Option.CONTEXT_SENSITIVITY_INFERENCE,
                ExperimentalOptions.Option.PROPERTY_CONTEXT_SENSITIVE_OBJECT_LITERALS,
                ExperimentalOptions.Option.SYNTACTIC_LOOP_VARIABLE_IDENTIFICATION
        );

        OptionValues options = new OptionValues();

        options.enableIncludeDom();
        options.enableQuiet();
        options.enableForInSpecialization();
        options.enableTest();
        options.enableUnevalizer();
        options.enableContextSensitiveHeap();
        options.enableParameterSensitivity();
        options.enableNumericVariableSensitivity();
        options.enableUnrollOneAndAHalf();
        return Pair.make(options, experimentalOptions);
    }

    private static void printUsage () {

        final String javaPrefix = "java -cp tajs-all.jar dk.brics.tajs.MainJQueryDeterminacy";
        System.out.printf("Usage: %s BENCHMARK MODE SOURCE%n", javaPrefix);
        System.out.printf(" where BENCHMARK is one of %s%n", Arrays.asList(Benchmark.values()));
        System.out.printf(" where MODE is one of %s%n", Arrays.asList(Mode.values()));
        System.out.printf(" where SOURCE is the JavaScript or HTML file to analyze%n");
        System.out.println();
        System.out.printf(" example 1 for benchmark group %s: %s %s %s benchmarks/jquery-1.10.js%n", Benchmark.A, javaPrefix, Benchmark.A, Mode.ALL);
        System.out.printf(" example 2 for benchmark group %s: %s %s %s benchmarks/jquery-1.10.js%n", Benchmark.A, javaPrefix, Benchmark.A, Mode.NO_BRANCH_PRUNING);
        System.out.printf(" example for benchmark group %s: %s %s %s benchmarks/22.html%n", Benchmark.B, javaPrefix, Benchmark.B, Mode.ALL);
        System.out.printf(" example for benchmark group %s: %s %s %s benchmarks/22.html%n", Benchmark.C, javaPrefix, Benchmark.C, Mode.ALL);
        System.out.printf(" example 1 for benchmark group %s: %s %s %s benchmarks/delta-blue.js%n", Benchmark.D, javaPrefix, Benchmark.D, Mode.ALL);
        System.out.printf(" example 2 for benchmark group %s: %s %s %s benchmarks/delta-blue.js%n", Benchmark.D, javaPrefix, Benchmark.D, Mode.NONE);
        System.out.printf(" example 3 for benchmark group %s: %s %s %s benchmarks/aquarium.html%n", Benchmark.D, javaPrefix, Benchmark.D, Mode.ALL);
        System.out.printf(" example 4 for benchmark group %s: %s %s %s benchmarks/aquarium.html%n", Benchmark.D, javaPrefix, Benchmark.D, Mode.NONE);

    }
}
