package dk.brics.tajs.analysis;

import dk.brics.tajs.lattice.Value;
import dk.brics.tajs.util.AnalysisException;

import java.util.List;
import java.util.Map;

import static dk.brics.tajs.util.Collections.newMap;

/**
 * Representation of positional arguments to a function.
 * To be used in {@link dk.brics.tajs.analysis.Context}.
 */
public class ContextArguments {
    private final Value unknownArg;
    private final List<String> parameterNames;
    private final List<Value> arguments;

    /**
     * Represents that the function was invoked with an unknown number and order of arguments
     *
     * @param unknownArg all the arguments in a single value
     */
    public ContextArguments(Value unknownArg) {
        this.unknownArg = unknownArg;
        this.arguments = null;
        this.parameterNames = null;
    }

    /**
     * @return true iff the number and order of arguments to the function are unknown
     */
    public boolean isUnknown() {
        return unknownArg != null;
    }

    public String toString(boolean sourcesOnly) {
        if (isUnknown()) {
            return "UnknownArg(" + unknownArg + ")";
        }
        final Map<String, String> mapLike = newMap();
        int i = 0;
        for (Value value : arguments) {
            if (parameterNames.size() > i) {
                mapLike.put(parameterNames.get(i), value == null ? "-" : sourcesOnly ? toSourcesOnly(value) : value.toString());
            }
            i++;
        }
        for (; i < arguments.size(); i++) {
            Value value = arguments.get(i);
            mapLike.put("<" + i + ">", value == null ? "-" : sourcesOnly ? toSourcesOnly(value) : value.toString());
        }
        return mapLike.toString();
    }

    private String toSourcesOnly(Value value) {
        return value.getObjectSourceLocations().toString();
    }

    @Override
    public String toString() {
        return toString(false);
    }

    public String toSourcesOnlyString() {
        return toString(true);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final ContextArguments that = (ContextArguments) o;

        if (arguments != null ? !arguments.equals(that.arguments) : that.arguments != null) return false;
        if (parameterNames != null ? !parameterNames.equals(that.parameterNames) : that.parameterNames != null)
            return false;
        if (unknownArg != null ? !unknownArg.equals(that.unknownArg) : that.unknownArg != null) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = unknownArg != null ? unknownArg.hashCode() : 0;
        result = 31 * result + (parameterNames != null ? parameterNames.hashCode() : 0);
        result = 31 * result + (arguments != null ? arguments.hashCode() : 0);
        return result;
    }

    /**
     * @return the arguments the function was invoked with
     */
    public List<Value> getArguments() {
        if (arguments == null) {
            throw new AnalysisException("No arguments!");
        }
        return arguments;

    }

    /**
     * @return the unknownArg, if present
     */
    public Value getUnknownArg() {
        if (unknownArg == null) {
            throw new AnalysisException("No unknown argument!");
        }
        return unknownArg;
    }

    /**
     * Represents that the function was invoked with a known order and number of arguments (standard invocation or precise apply/call).
     *
     * @param parameterNames as the syntactically known parameter names of the invoked function (used for pretty printing)
     * @param arguments      as the arguments, in the order they are provided to the function
     */
    public ContextArguments(List<String> parameterNames, List<Value> arguments) {
        this.parameterNames = parameterNames;
        this.arguments = arguments;
        this.unknownArg = null;

    }
}
