/*
 * Copyright 2009-2013 Aarhus University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.brics.tajs.analysis.dom;

import dk.brics.tajs.analysis.InitialStateBuilder;
import dk.brics.tajs.analysis.State;
import dk.brics.tajs.analysis.dom.ajax.AjaxBuilder;
import dk.brics.tajs.analysis.dom.core.CoreBuilder;
import dk.brics.tajs.analysis.dom.core.DOMDocument;
import dk.brics.tajs.analysis.dom.event.EventBuilder;
import dk.brics.tajs.analysis.dom.html.HTMLBuilder;
import dk.brics.tajs.analysis.dom.html5.HTML5Builder;
import dk.brics.tajs.analysis.dom.style.CSSStyleDeclaration;
import dk.brics.tajs.analysis.dom.style.StyleBuilder;
import dk.brics.tajs.analysis.dom.view.ViewBuilder;
import dk.brics.tajs.lattice.BlockState;
import dk.brics.tajs.lattice.ObjectLabel;
import dk.brics.tajs.lattice.Value;

import java.util.Set;

import static dk.brics.tajs.analysis.dom.DOMFunctions.createDOMProperty;
import static dk.brics.tajs.util.Collections.newSet;

/**
 * Setup the DOM browser model.
 * <p/>
 * An overview is available at:
 * <p/>
 * http://dsmith77.files.wordpress.com/2008/07/the-document-object-model-dom.gif
 * <p/>
 * DOM Spec:
 * http://www.w3.org/DOM/DOMTR
 * <p/>
 * DOM LEVEL 1:
 * http://www.w3.org/TR/2000/WD-DOM-Level-1-20000929/ecma-script-language-binding.html
 * <p/>
 * DOM Level 2 Core:
 * http://www.w3.org/TR/DOM-Level-2-Core/core.html
 * http://www.w3.org/TR/DOM-Level-2-Core/ecma-script-binding.html
 * <p/>
 * DOM LEVEL 2 HTML:
 * http://www.w3.org/TR/DOM-Level-2-HTML/ecma-script-binding.html
 * <p/>
 * DOM LEVEL 2: Traversal
 * http://www.w3.org/TR/DOM-Level-2-Traversal-Range/Overview.html
 */
public class DOMBuilder {
    /**
     * Construct the initial DOM objects.
     * Its assumed that WINDOW is added to the state somewhere else before this function is invoked since its the
     * global objects when running in DOM mode.
     */
    public static void addInitialState(State s) {
        // Reset DOM Registry
        DOMRegistry.reset();

        // Build window object
        DOMWindow.WINDOW = InitialStateBuilder.GLOBAL;
        DOMWindow.build(s);

        // Build initial core dom state
        CoreBuilder.build(s);

        // Build initial style state
        StyleBuilder.build(s);

        // Build initial html state
        HTMLBuilder.build(s);

        // Build initial html5 state
        HTML5Builder.build(s);

        // Build initial event state
        EventBuilder.build(s);

        // Build initial views state
        ViewBuilder.build(s);

        // Build initial AJAX state
        AjaxBuilder.build(s);


        // Set some shared properties on DOM elements due to circularity, and convenience
        Set<ObjectLabel> htmlElements = newSet();
        htmlElements.addAll(HTMLBuilder.HTML_OBJECT_LABELS);
        htmlElements.addAll(HTML5Builder.HTML5_OBJECT_LABELS);
        Value cssProperty = Value.makeObject(CSSStyleDeclaration.STYLEDECLARATION).setReadOnly();
        Value htmlElementsProperty = Value.makeObject(htmlElements).joinNull().setReadOnly();
        Value uintProperty = Value.makeAnyNumUInt().setReadOnly();
        BlockState.I_KNOW_WHAT_I_AM_DOING_DURING_INITIALIZATION = true;
        for (ObjectLabel element : htmlElements) {
            createDOMProperty(s, element, "clientWidth", uintProperty);
            createDOMProperty(s, element, "clientHeight", uintProperty);
            createDOMProperty(s, element, "scrollWidth", uintProperty);
            createDOMProperty(s, element, "scrollHeight", uintProperty);
            createDOMProperty(s, element, "style", cssProperty);
            createDOMProperty(s, element, "firstChild", htmlElementsProperty);
            createDOMProperty(s, element, "parentNode", htmlElementsProperty);
            createDOMProperty(s, element, "lastChild", htmlElementsProperty);
            createDOMProperty(s, element, "previousSibling", htmlElementsProperty);
            createDOMProperty(s, element, "nextSibling", htmlElementsProperty);
        }
        Set<ObjectLabel> htmlElementsAndWindow = newSet(htmlElements);
        htmlElementsAndWindow.add(DOMWindow.WINDOW);
        for (ObjectLabel element : htmlElementsAndWindow) {
            createDOMProperty(s, element, "onsubmit", Value.makeNull());
            createDOMProperty(s, element, "onchange", Value.makeNull());
        }
        createDOMProperty(s, DOMDocument.INSTANCES, "defaultView", Value.makeObject(DOMWindow.WINDOW).joinNull().setReadOnly());
        BlockState.I_KNOW_WHAT_I_AM_DOING_DURING_INITIALIZATION = false;
        DOMFunctions.makeAnyHTMLNodeList(s);
    }
}
