/*
 * Copyright 2009-2013 Aarhus University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.brics.tajs.analysis.nativeobjects;

import dk.brics.tajs.analysis.*;
import dk.brics.tajs.analysis.FunctionCalls.CallInfo;
import dk.brics.tajs.analysis.dom.DOMFunctions;
import dk.brics.tajs.analysis.dom.ajax.ReadystateEvent;
import dk.brics.tajs.analysis.dom.event.*;
import dk.brics.tajs.analysis.uneval.NormalForm;
import dk.brics.tajs.analysis.uneval.UnevalTools;
import dk.brics.tajs.flowgraph.AbstractNode;
import dk.brics.tajs.flowgraph.FlowGraph;
import dk.brics.tajs.flowgraph.FlowGraphFragment;
import dk.brics.tajs.flowgraph.Function;
import dk.brics.tajs.flowgraph.jsnodes.CallNode;
import dk.brics.tajs.js2flowgraph.RhinoASTToFlowgraph;
import dk.brics.tajs.lattice.BlockState;
import dk.brics.tajs.lattice.ObjectLabel;
import dk.brics.tajs.lattice.ObjectLabel.Kind;
import dk.brics.tajs.lattice.UnknownValueResolver;
import dk.brics.tajs.lattice.Value;
import dk.brics.tajs.options.ExperimentalOptions;
import dk.brics.tajs.options.Options;
import dk.brics.tajs.solver.Message.Severity;
import dk.brics.tajs.solver.NodeAndContext;
import dk.brics.tajs.unevalizer.Unevalizer;
import dk.brics.tajs.util.AnalysisException;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static dk.brics.tajs.util.Collections.newSet;

/**
 * 15.1 and B.2 native global functions.
 */
public class JSGlobal {
    public static final Set<ECMAScriptObjects>TAJS_HOOK_FUNCTIONS = new HashSet<>();
    static{
        TAJS_HOOK_FUNCTIONS.addAll(Arrays.asList(new ECMAScriptObjects[]{ECMAScriptObjects.ASSERT, ECMAScriptObjects.DUMPVALUE, ECMAScriptObjects.DUMPPROTOTYPE, ECMAScriptObjects.DUMPOBJECT, ECMAScriptObjects.DUMPSTATE, ECMAScriptObjects.DUMPMODIFIEDSTATE, ECMAScriptObjects.DUMPATTRIBUTES, ECMAScriptObjects.DUMPOBJECTORIGIN, ECMAScriptObjects.DUMPEXPRESSION, ECMAScriptObjects.DUMPNF, ECMAScriptObjects.ASSERT_SINGLE_NUM, ECMAScriptObjects.ASSERT_SINGLE_STR, ECMAScriptObjects.TAJS_ASSERT, ECMAScriptObjects.TAJS_CONVERSION_TO_PRIMITIVE, ECMAScriptObjects.ASSUME_VALUE, ECMAScriptObjects.ASSUME_TYPEOF, ECMAScriptObjects.TAJS_ADD_CONTEXT_SENSITIVITY, ECMAScriptObjects.TAJS_NEW_OBJECT, ECMAScriptObjects.TAJS_DUMP_CONTEXT_ARGUMENTS}));
    }

	private static Logger logger = Logger.getLogger(JSGlobal.class); 

	private JSGlobal() {}

	/**
	 * Evaluates the given native function.
	 */
	public static Value evaluate(ECMAScriptObjects nativeobject, CallInfo call, final State state, Solver.SolverInterface c) {
		if (NativeFunctions.throwTypeErrorIfConstructor(call, state, c))
			return Value.makeNone();

		switch (nativeobject) {

		case EVAL: { // 15.1.2.1
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value evalValue = NativeFunctions.readParameter(call, state, 0);
			if (evalValue.isMaybeOtherThanStr()) {
				if (!evalValue.isNotStr()) {
                    String msg = "Parameter to eval is maybe a string and maybe a non-string, we currently can't handle that, sorry";
                    if (Options.isUnsoundEnabled()) {
                        c.getMonitoring().addMessage(call.getJSSourceNode(), Severity.TAJS_META, "Unsound ignore: " + msg);
                        return Value.makeNone();
                    }
                    throw new AnalysisException(msg); // TODO: improve?
                }
				// TODO: issue warning if calling eval with non-string value
				return evalValue;
			}
            if (Options.isEvalStatistics()) 
            	c.getMonitoring().visitEvalCall(call.getSourceNode(), NativeFunctions.readParameter(call, state, 0));
			if (c.isScanning())
				return Value.makeNone();
			if (evalValue.isStrJSON()) { 
				return DOMFunctions.makeAnyJSONObject(state).join(evalValue.restrictToNotStr());
			} else if (Options.isUnevalizerEnabled()) {
                CallNode evalCall = (CallNode) call.getSourceNode();
                FlowGraph currentFg = c.getFlowGraph();
                boolean ignoreResult = evalCall.getResultRegister() == AbstractNode.NO_VALUE;
                String var = ignoreResult ? null : UnevalTools.gensym(); // Do we need the value of the eval call after?
                NormalForm input = UnevalTools.rebuildNormalForm(currentFg, evalCall, state, c);

                // Collect special args that should be analyzed context sensitively
                c.getAnalysis().getSpecialVars().addContextSensitivity(evalCall.getBlock().getFunction(), input.getArgumentsInUse(), state, c);

                // What we should use as key for the eval cache is the entire tuple from the Uneval paper. Since that
                // might contain infinite sets and other large things we just call the Unevalizer and compare the output
                // of the Unevalizer to the key in the cache. This makes us Uneval more things, but we save the work
                // of re-extending the flow graph every time.
                boolean aliased_call = !"eval".equals(UnevalTools.get_call_name(currentFg, evalCall)); // TODO: aliased_call should also affect the execution context?
                String unevaled = new Unevalizer().uneval(UnevalTools.unevalizerCallback(currentFg, state, evalCall, input), input.getNormalForm(), aliased_call, var);

                if (unevaled == null) {
                    String msg = "Unevalable eval: " + UnevalTools.rebuildFullExpression(currentFg, evalCall, evalCall.getArgRegister(0));
                    if (Options.isUnsoundEnabled()) {
                        c.getMonitoring().addMessage(call.getJSSourceNode(), Severity.TAJS_META, "Unsound ignore: " + msg);
                        return Value.makeNone();
                    }
                    throw new AnalysisException(msg);
                }
                if (logger.isDebugEnabled())
    				logger.debug("Unevalized: " + unevaled);

               	unevaled = UnevalTools.rebuildFullFromMapping(currentFg, unevaled, input.getMapping(), evalCall);

                String unevaledSubst = ignoreResult ? unevaled : unevaled.replace(var, UnevalTools.VAR_PLACEHOLDER); // to avoid the random string in the cache
                EvalCache evalCache = c.getAnalysis().getEvalCache();
                NodeAndContext<Context> cc = new NodeAndContext<>(evalCall, state.getContext());
                FlowGraphFragment e = evalCache.getCode(cc);

                // Cache miss.
                if (e == null || !e.getKey().equals(unevaledSubst))
                    e = (new RhinoASTToFlowgraph()).extendFlowgraph(currentFg, unevaled, unevaledSubst, evalCall, e, var);

                evalCache.setCode(cc, e);
                c.propagateToBasicBlock(state.clone(), e.getEntryBlock(), state.getContext());
                if (Options.isFlowGraphEnabled()) {
                	try (PrintWriter pw = new PrintWriter(new File("out" + File.separator + "flowgraphs" + File.separator + "uneval-" + 
                			evalCall.getIndex() + "-" + Integer.toHexString(state.getContext().hashCode()) + ".dot"))) {
                		currentFg.toDot(pw);
                        pw.flush();
                    } catch (Exception ee) {
                        throw new AnalysisException(ee);
                    }
                }
                return Value.makeNone();
			} else {
				throw new AnalysisException("eval of non JSONStr not supported, and unevalizer is not enabled");
			}
		}
		case PARSEINT: { // 15.1.2.2
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 2);
			Value str = Conversion.toString(NativeFunctions.readParameter(call, state, 0), c);
			Value basis;
			if (call.isUnknownNumberOfArgs()) {
				basis = NativeFunctions.readParameter(call, state, 1).joinNum(0);
			} else {
				basis = call.getNumberOfArgs() >= 2 ? Conversion.toNumber(NativeFunctions.readParameter(call, state, 1), c) : Value.makeNum(0);
			}
			if (str.isMaybeSingleStr() && basis.isMaybeSingleNum()) {
				String s = str.getStr().trim();
				double sign = 1;
				if (s.startsWith("-"))
					sign = -1;
				if (s.startsWith("-") || s.startsWith("+"))
					s = s.substring(1);
				int radix = Conversion.toInt32(basis.getNum());
				if (radix == 0) {
					radix = 10;
					if (s.length() > 1 && s.startsWith("0")) {
						radix = 8;
						if (s.length() > 2 && Character.toLowerCase(s.charAt(1)) == 'x') {
							radix = 16;
							s = s.substring(2);
						}
					}
				}
				if (radix < 2 || radix > 36)
					return Value.makeNum(Double.NaN);
				else {
					int i;
					String z = s;
					for (i = 0; i < s.length(); i++)
						if (Character.digit(s.charAt(i), radix) < 0) {
							z = s.substring(0, i);
							break;
						}
					if (z.isEmpty())
						return Value.makeNum(Double.NaN);
					else
						return Value.makeNum(sign * Integer.parseInt(z, radix));
				}
			} else
				return Value.makeAnyNum();
		}

		case PARSEFLOAT: { // 15.1.2.3
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value str = Conversion.toString(NativeFunctions.readParameter(call, state, 0), c);
			if (str.isMaybeSingleStr()) {
				String s = str.getStr().trim();
				Pattern p = Pattern.compile("[+-]?(Infinity|([0-9]+\\.[0-9]*|\\.[0-9]+|[0-9]+)([eE][+-]?[0-9]+)?)");
				Matcher m = p.matcher(s);
				if (m.lookingAt())
					return Value.makeNum(Double.parseDouble(m.group(0)));
				else
					return Value.makeNum(Double.NaN);
			} else
				return Value.makeAnyNum();
		}

		case ISNAN: { // 15.1.2.4
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value num = Conversion.toNumber(NativeFunctions.readParameter(call, state, 0), c);
			Value res = Value.makeNone();
			if (num.isMaybeNaN())
				res = res.joinBool(true);
			if (num.isMaybeSingleNum() || num.isMaybeInf() || num.isMaybeNumUInt() || num.isMaybeNumOther())
				res = res.joinBool(false);
			return res; 
		}

		case ISFINITE: { // 15.1.2.5
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value num = Conversion.toNumber(NativeFunctions.readParameter(call, state, 0), c);
			if (num.isMaybeSingleNum())
				return Value.makeBool(!num.getNum().isInfinite());
			Value res = Value.makeNone();
			if (num.isMaybeNaN() || num.isMaybeInf())
				res = res.joinBool(false);
			if (num.isMaybeNumUInt() || num.isMaybeNumOther())
				res = res.joinBool(true);
			return res; 
		}

		case PRINT:  // in Rhino, expects any number of parameters; returns undefined
		case ALERT: {
			return Value.makeUndef();
		}

		case DECODEURI: // 15.1.3.1
		case DECODEURICOMPONENT: // 15.1.3.2
		case ENCODEURI: // 15.1.3.3
		case ENCODEURICOMPONENT: // 15.1.3.4
		case ESCAPE: // B.2.1
		case UNESCAPE: { // B.2.2
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
            /* Value arg = */ NativeFunctions.readParameter(call, state, 0);
			return Value.makeAnyStr(); // TODO: could improve precision for constant strings
		}

		case DUMPVALUE: {
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value x = NativeFunctions.readParameter(call, state, 0); // to avoid recover: call.getArg(0);
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Abstract value: " + x.restrictToNotModified() /*+ " (context: " + c.getCurrentContext() + ")"*/);
			return Value.makeUndef();
		}

		case DUMPPROTOTYPE: {
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value x = NativeFunctions.readParameter(call, state, 0);

			StringBuilder sb = new StringBuilder();
			Value p = state.readInternalPrototype(x.getObjectLabels());
			p = UnknownValueResolver.getRealValue(p, state);
			while (p.isMaybeObject()) {
				sb.append(p.toString());
				p = state.readInternalPrototype(p.getObjectLabels());
				p = UnknownValueResolver.getRealValue(p, state);
				if (!p.isNullOrUndef()) {
					sb.append(" -> ");
				}
			}

			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Prototype: " + sb);
			return Value.makeUndef();
		}

		case DUMPOBJECT: {
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value x = NativeFunctions.readParameter(call, state, 0);
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Abstract object: " + state.printObject(x) /*+ " (context: " + c.getCurrentContext() + ")"*/);
			return Value.makeUndef();
		}

		case DUMPSTATE: {
			NativeFunctions.expectParameters(nativeobject, call, c, 0, 0);
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Abstract state:\n" + state /*+ " (context: " + c.getCurrentContext() + ")"*/);
			/*
			try {
                File outdir = new File("out");
                if (!outdir.exists()) {
                    outdir.mkdir();
                }
				FileWriter fw = new FileWriter("out" + File.separator + "state.dot");
				fw.write(state.toDot());
				fw.close();			
			} catch (IOException e) {
				throw new AnalysisException(e);
			}
			*/
			return Value.makeUndef();
		}

		case DUMPMODIFIEDSTATE: {
			NativeFunctions.expectParameters(nativeobject, call, c, 0, 0);
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Abstract state (modified parts):" /*+ " (context: " + c.getCurrentContext() + ")"*/ + state.toStringModified());
			return Value.makeUndef();
		}

		case DUMPATTRIBUTES: {
			NativeFunctions.expectParameters(nativeobject, call, c, 2, 2);
			Value x = NativeFunctions.readParameter(call, state, 0);
			Value p = Conversion.toString(NativeFunctions.readParameter(call, state, 1), c);
			if (!p.isMaybeSingleStr())
				c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling dumpAttributes with non-constant property name");
			else {
				String propertyname = p.getStr();
				Value v = state.readPropertyDirect(x.getObjectLabels(), propertyname);
				c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Property attributes: " + v.printAttributes() /*+ " (context: " + c.getCurrentContext() + ")"*/);
			}
			return Value.makeUndef();
		}

		case DUMPOBJECTORIGIN: { // TODO: remove dumpObjectOrigin? (use dumpObject instead)
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			Value x = NativeFunctions.readParameter(call, state, 0);
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Origin of objects: " + BlockState.printObjectOrigins(x) /*+ " (context: " + c.getCurrentContext() + ")"*/);
			return Value.makeUndef();
		}

		case DUMPEXPRESSION: {
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			CallNode cn =(CallNode) call.getSourceNode();
			String s = UnevalTools.rebuildFullExpression(c.getFlowGraph(), call.getSourceNode(), cn.getArgRegister(0));
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Exp: " + s);
			return Value.makeUndef();
		}

		case DUMPNF: {
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
			CallNode cn =(CallNode) call.getSourceNode();
			NormalForm s = UnevalTools.rebuildNormalForm(c.getFlowGraph(), cn, state, c);
			c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "NF: " + s.toString());
			return Value.makeUndef();

		}

		case TAJS_CONVERSION_TO_PRIMITIVE: {
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 2);
			Value varg = NativeFunctions.readParameter(call, state, 0);
			Value vhint;
			if (call.isUnknownNumberOfArgs())
				vhint = NativeFunctions.readParameter(call, state, 1).joinStr("NONE");
			else
				vhint = call.getNumberOfArgs() >= 2 ? NativeFunctions.readParameter(call, state, 1) : Value.makeStr("NONE");
				if (!vhint.isMaybeSingleStr()) {
					c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling conversionToPrimitive with non-constant hint string");
					return Value.makeUndef();
				} else {
					String shint = vhint.getStr();
					return Conversion.toPrimitive(varg, shint.equals("NONE") ? Conversion.Hint.NONE : shint.equals("NUM") ? Conversion.Hint.NUM : Conversion.Hint.STR, c);
				}
		}


            case TAJS_ADD_CONTEXT_SENSITIVITY: {
            if (!Options.isParameterSensitivityEnabled() || ExperimentalOptions.ExperimentalOptionsManager.get().isEnabled(ExperimentalOptions.Option.CONTEXT_SENSITIVITY_INFERENCE)) {
                return Value.makeUndef();
            }
            if (call.isUnknownNumberOfArgs())
                    throw new AnalysisException("Calling TAJS_addContextSensitivity with unknown number of arguments");
                Value param;
                Set<Function> functions = newSet();
                if (call.getNumberOfArgs() == 1) { // enclosing function
                    param = NativeFunctions.readParameter(call, state, 0);
                    functions.add(call.getSourceNode().getBlock().getFunction());
                } else if (call.getNumberOfArgs() == 2) { // function given as first parameter
                    Value funval = NativeFunctions.readParameter(call, state, 0);
                    Set<ObjectLabel> objlabels = funval.getObjectLabels();
                    for (ObjectLabel objlabel : objlabels) {
                        if (objlabel.getKind() == Kind.FUNCTION && !objlabel.isHostObject()){
                            functions.add(objlabel.getFunction());
                        }
                    }
                    if (functions.isEmpty()) {
                        c.getMonitoring().addMessage(call.getSourceNode(), Severity.TAJS_META, "Calling TAJS_addContextSensitivity without user-function argument: " + funval);
                    }
                    param = NativeFunctions.readParameter(call, state, 1);
                } else {
                    throw new AnalysisException("Calling TAJS_addContextSensitivity with unexpected number of arguments");
                }

                String var;
                if (param.isMaybeSingleStr()) {
                    if (param.isMaybeOtherThanStr())
                        throw new AnalysisException("Calling TAJS_addContextSensitivity with non-fixed-string argument: " + param);
                    var = param.getStr();
                    for (Function function : functions) {
                        if (!function.getParameterNames().contains(param) &&
                            !Options.isNumericVariableSensitivityEnabled()) {
                                continue;
                            }
                        c.getAnalysis().getSpecialVars().addToSpecialVars(function, var);
                    }
                } else if (param.isMaybeSingleNum()) {
                    if (param.isMaybeOtherThanNumUInt())
                        throw new AnalysisException("Calling TAJS_addContextSensitivity with non-fixed-index argument: " + param);
                    for (Function function : functions) {
                        List<String> parameterNames = function.getParameterNames();
                        int index = param.getNum().intValue();
                        if (parameterNames.size() > index) {
                            var = parameterNames.get(index);
                            c.getAnalysis().getSpecialVars().addToSpecialVars(function, var);
                        } else {
                            c.getMonitoring().addMessage(call.getSourceNode(), Severity.TAJS_META, String.format("Calling TAJS_addContextSensitivity with index for non-existing parameter: %s (for %s)", index, function.getSourceLocation()));
                        }
                    }
                } else {
                    throw new AnalysisException("Calling TAJS_addContextSensitivity with unexpected argument: " + param);
                }
                return Value.makeUndef();
            }
        case TAJS_DUMP_CONTEXT_ARGUMENTS: {
            NativeFunctions.expectParameters(nativeobject, call, c, 0, 0);
            Context context = state.getContext();
            if (context != null) {
                ContextArguments arguments = context.getEntryBlockAndContext().getContext().getArguments();
                final String msg = String.format("Arguments: %s (%s)", arguments.toString(), arguments.toSourcesOnlyString());
                c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.TAJS_META, msg);
            }
            return Value.makeUndef();
        }
		case TAJS_NEW_OBJECT: {
			ObjectLabel objlabel;
            if (Options.isContextSensitiveHeapEnabled()) {
                objlabel = new ObjectLabel(call.getSourceNode(), Kind.OBJECT,
                        state.getContext().getArguments(), state.getContext().getSpecialVars(), state.getContext().getSpecialRegisters());
            } else {
                objlabel = new ObjectLabel(call.getSourceNode(), Kind.OBJECT);
            }
 			state.newObject(objlabel);
 	        state.writeInternalPrototype(objlabel, Value.makeObject(InitialStateBuilder.OBJECT_PROTOTYPE));
         	return Value.makeObject(objlabel);
		}
        case ASSUME_VALUE:{ // could be implemented with a simple assignment
            if(!ExperimentalOptions.ExperimentalOptionsManager.get().isEnabled(ExperimentalOptions.Option.MANUAL_ASSUMES)){
                return Value.makeUndef();
            }
            // TODO support multiplevalues for joins..
            NativeFunctions.expectParameters(nativeobject, call, c, 2, 2);
            if (call.isUnknownNumberOfArgs()) {
                c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling assumeValue with unknown number of arguments");
            } else {
                Value varg = NativeFunctions.readParameter(call, state, 0);
                if (!varg.isMaybeSingleStr()) {
                    c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling assumeValue with non-constant variable string");
                } else {
                    String varname = varg.getStr();
                    Value v = state.readVariable(varname, null);
                    Value restricted = NativeFunctions.readParameter(call, state, 1);
                    state.writeVariable(varname, restricted, false);
                }
            }
            return Value.makeUndef();
        }

        case ASSUME_TYPEOF: {
            if(!ExperimentalOptions.ExperimentalOptionsManager.get().isEnabled(ExperimentalOptions.Option.MANUAL_ASSUMES)){
                return Value.makeUndef();
            }
            // assumeTypeOf(variableName, typeof-string, comparisonResult)
            // e.x. assumeTypeOf('x', 'object', true)
            // restricts the given variable value to be of a type compatible with a typeof check: e.g. 'typeof x === "object"' => x is restricted to be an object
            // third argument is the expected result of the typeof comparison. If false, then the comparison which is assumed yields false, e.g. 'typeof x !== "object"'
            NativeFunctions.expectParameters(nativeobject, call, c, 3, 3);
            if (call.isUnknownNumberOfArgs()) {
                c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling assumeTypeOf with unknown number of arguments");
            } else {
                Value var = NativeFunctions.readParameter(call, state, 0);
                Value type = NativeFunctions.readParameter(call, state, 1);
                Value expectedResult = NativeFunctions.readParameter(call, state, 2);
                if (!var.isMaybeSingleStr()) {
                    c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling assumeTypeOf with non-constant variable string: " + var);
                } else if (!type.isMaybeSingleStr()) {
                    c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling assumeTypeOf with non-constant type string " + type);
                } else if (!(expectedResult.isMaybeTrueButNotFalse() || expectedResult.isMaybeFalseButNotTrue())) {
                    c.getMonitoring().addMessageInfo(c.getCurrentNode(), Severity.HIGH, "Calling assumeTypeOf with non-constant boolean comparisonResult: " + expectedResult);
                } else {
                    String varName = var.getStr();
                    Set<Value> restricteds = newSet();
                    Value v = state.readVariable(varName, null);
                    v = UnknownValueResolver.getRealValue(v, state);
                    for (String typeName : type.getStr().split("\\|\\|")) {
                        typeName = typeName.trim();
                        Value restricted = null;
                        switch (typeName) {
                            case "undefined":
                                restricted = v.restrictToUndef();
                                break;
                            case "object": {
                                Set<ObjectLabel> objectLabels = v.getObjectLabels();
                                Set<ObjectLabel> nonFunctionObjects = newSet();
                                for (ObjectLabel l : objectLabels) {
                                    if (l.getKind() != Kind.FUNCTION)
                                        nonFunctionObjects.add(l);
                                }
                                restricted = Value.makeObject(nonFunctionObjects);
                                break;
                            }
                            case "boolean":
                                restricted = v.restrictToBool();
                                break;
                            case "number":
                                restricted = v.restrictToNum();
                                break;
                            case "string":
                                restricted = v.restrictToStr();
                                break;
                            case "function": {
                                Set<ObjectLabel> objectLabels = v.getObjectLabels();
                                Set<ObjectLabel> functionObjects = newSet();
                                for (ObjectLabel l : objectLabels) {
                                    if (l.getKind() == Kind.FUNCTION)
                                        functionObjects.add(l);
                                }
                                restricted = Value.makeObject(functionObjects);
                                break;
                            }
                            case "xml":
                                // TODO handle assumeTypeOf(x,"xml")
                            default:
                                throw new AnalysisException("Calling assumeTypeOf with unsupported type: '" + typeName + "'");
                        }
                        assert (restricted != null);
                        restricteds.add(restricted);
                    }
                    Value restricted = Value.join(restricteds);
                    if (expectedResult.isMaybeFalseButNotTrue()) {
                        // typeof check expected to fail. So remove successful restriction from original value instead!
                        restricted = v.remove(restricted);
                    }
                    if (restricted.isNotPresent())
                        return Value.makeNone();
                    state.writeVariable(varName, restricted, false);
                }

            }
            return Value.makeUndef();
        }

            case TAJS_GET_UI_EVENT: {
			return Value.makeObject(UIEvent.INSTANCES);
		}

		case TAJS_GET_MOUSE_EVENT: {
			return Value.makeObject(MouseEvent.INSTANCES);
		}

		case TAJS_GET_AJAX_EVENT: {
			return Value.makeObject(ReadystateEvent.INSTANCES);
		}

		case TAJS_GET_KEYBOARD_EVENT: {
			return Value.makeObject(KeyboardEvent.INSTANCES);
		}

		case TAJS_GET_EVENT_LISTENER: {
			return Value.makeObject(EventListener.INSTANCES);
		}

		case TAJS_GET_WHEEL_EVENT: {
			return Value.makeObject(WheelEvent.INSTANCES);
		}

		case ASSERT: { // TODO: make this a special case of TAJS_assert?
			NativeFunctions.expectParameters(nativeobject, call, c, 1, 2);
            Value x = Conversion.toBoolean(NativeFunctions.readParameter(call, state, 0));
            if (x.isMaybeFalse()) {
            	if (Conversion.toBoolean(NativeFunctions.readParameter(call, state, 1)).isMaybeTrue()) { // used by TestJSNativeFunctions
            		throw new AnalysisException("Assertion fails!");
            	}
            	c.getMonitoring().addMessage(c.getCurrentNode(), Severity.HIGH, "Assertion fails");
            }
			return Value.makeUndef();
		}

        case ASSERT_SINGLE_NUM: { // TODO: use TAJS_assert instead?
            NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
            Value arg = NativeFunctions.readParameter(call, state, 0);
            if(!arg.isMaybeSingleNum()) {
                throw new AnalysisException("Assertion failed:  Not a SingleNum: " + arg.toString());
            }
            return Value.makeUndef();
        }

        case ASSERT_SINGLE_STR: { // TODO: use TAJS_assert instead?
            NativeFunctions.expectParameters(nativeobject, call, c, 1, 1);
            Value arg = NativeFunctions.readParameter(call, state, 0);
            if (!arg.isMaybeSingleStr()) {
                throw new AnalysisException("Assertion failed:  Not a SingleStr: " + arg.toString());
            }
            return Value.makeUndef();
        }


            case TAJS_ASSERT: {
            return tajsValueAssert(nativeobject, call, state, c);
        }
		default:
			return null;
		}
	}

    /**
     * Generic value assertion. Could replace all other asserts.
     * <p/>
     * Signature:
     * <p/>
     * <pre>
     * _TAJS_assert(value, predicate, expectedResult)
     * </pre>
     * <p/>
     * The predicate is an expression of predicate method names on {@link Value}, the method names can be combined with the logical connector '||', whitespace is ignored.
     * <p/>
     * Examples:
     * <pre>
     * _TAJS_assert(value, isMaybeAnyNum, expectedResult)
     * _TAJS_assert(value, isMaybeAnyNum || isMaybeAnyString, expectedResult)
     * _TAJS_assert(value, isMaybeAnyNum || isMaybeAnyString || isMaybeObject, expectedResult)
     * </pre>
     * <p/>
     * where:
     * <ul>
     * <li><b>value</b> is the JavaScript value to assert something about</li>
     * <li><b>predicate</b> is the predicates to evaluate <b>value</b> with</li>
     * <li><b>expectedResult</b> is the single boolean expected return value of the predicate invocation. E.g. true</li>
     * </ul>
     */
    private static Value tajsValueAssert(ECMAScriptObjects nativeobject, CallInfo call, final State state, Solver.SolverInterface c) {
        NativeFunctions.expectParameters(nativeobject, call, c, 3, 3);
        Value value = NativeFunctions.readParameter(call, state, 0);
        Value predicate = NativeFunctions.readParameter(call, state, 1);
        Value expectedResult = NativeFunctions.readParameter(call, state, 2);
        if (!predicate.isMaybeSingleStr()) {
            throw new AnalysisException(String.format("Call to %s failed. Predicate '%s' is not a single string!", ECMAScriptObjects.TAJS_ASSERT.toString(), predicate));
        }
        if (!(expectedResult.isMaybeTrueButNotFalse() || expectedResult.isMaybeFalseButNotTrue())) {
            throw new AnalysisException(String.format("Call to %s failed. Expected result '%s' is not a single boolean!", ECMAScriptObjects.TAJS_ASSERT.toString(), expectedResult));
        }
        final String predicateString = predicate.getStr().replace(" ", "");
        boolean result = false;
        for (String methodName : predicateString.split("\\|\\|")) {
            try {
                Method method = Value.class.getMethod(methodName.trim());
                result = result || (boolean) method.invoke(value);
            } catch (NoSuchMethodException | SecurityException | IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
                throw new AnalysisException(String.format("Call to %s failed (%s). Predicate '%s' is not a zero-argument predicate on the Value class!", ECMAScriptObjects.TAJS_ASSERT.toString(), e.getClass().getSimpleName(), methodName));
            }
        }
        if ((expectedResult.isMaybeTrueButNotFalse() && !result) || (expectedResult.isMaybeFalseButNotTrue() && result)) {
            throw new AnalysisException(String.format("Assertion failed:  <%s>(%s) != %s (%s)", predicate, value, expectedResult, call.getSourceNode().getSourceLocation()));
        }

        return Value.makeUndef();
    }
}
