package dk.brics.tajs.analysis.nativeobjects.concrete;

import dk.brics.tajs.util.None;
import dk.brics.tajs.util.Optional;
import dk.brics.tajs.util.Some;
import org.apache.log4j.Logger;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.NativeArray;
import org.mozilla.javascript.ScriptableObject;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import java.util.ArrayList;
import java.util.List;


/**
 * Concrete semantics implementation.
 * Will perform concrete evaluation of concrete values using the Java native implementation of JavaScript.
 */
public class ConcreteSemantics {
    private static final Logger log = Logger.getLogger(ConcreteSemantics.class);
    private static ConcreteSemantics instance;
    private final ScriptEngineManager factory;

    private ConcreteSemantics() {
        this.factory = new ScriptEngineManager();
    }

    public static ConcreteSemantics get() {
        if (instance == null) {
            instance = new ConcreteSemantics();
        }
        return instance;
    }

    public static void main(String[] args) throws ScriptException {
        System.out.println(get().toConcreteValue((get().makeEngine().eval("[null, undefined]"))));
    }


    @SuppressWarnings("unchecked")
    public <T extends ConcreteValue> Optional<T> apply(String functionName, ConcreteValue base, List<ConcreteValue> arguments) {
        ScriptEngine engine = makeEngine();
        String script = String.format("%s.apply(%s, [%s]);", functionName, base.toSourceCode(), makeList(arguments));
        // System.out.println(script);
        try {
            Context cx = Context.enter();
            ScriptableObject scope = cx.initStandardObjects();
            Object result = cx.evaluateString(scope, script, "<concrete-execution>", 1, null);

            //Object result = engine.eval(script);
            // System.out.println("   ==> " + result);
            //System.out.println(scripts.getResults().size());
            return (Optional<T>) Some.make(toConcreteValue(result));
        } catch (Throwable t) {
            log.error(t);
            return None.make();
        }
    }

    private ScriptEngine makeEngine() {
        return factory.getEngineByName("JavaScript");
    }

    private static String makeList(List<ConcreteValue> arguments) {
        StringBuilder sb = new StringBuilder();
        boolean first = true;
        for (ConcreteValue argument : arguments) {
            if (!first) {
                sb.append(", ");
            }
            first = false;
            sb.append(argument.toSourceCode());
        }

        return sb.toString();
    }

    private ConcreteArray toConcreteArray(NativeArray array) { // TODO: cannot find javadoc for sun.org.mozilla.javascript.internal.NativeArray
        final ArrayList<ConcreteValue> concreteEntries = new ArrayList<>();
        for (Object entry : array) {
            concreteEntries.add(toConcreteValue(entry));
        }
        return new ConcreteArray(concreteEntries);
    }

    private ConcreteValue toConcreteValue(Object value) {
        if (value == null) {
            return new ConcreteNullOrUndefined(); // "Bug/feature" in ScriptEngine...
        }
        if (value instanceof NativeArray) {
            return toConcreteArray((NativeArray) value);
        }
        if (value instanceof Number) {
            return new ConcreteNumber(((Number) value).doubleValue());
        }
        if (value instanceof String) {
            return new ConcreteString(((String) value).replace("\n", "\\n").replace("\r", "\\r").replace("\t", "\\t").replace("\n", "\\n").replace("\f", "\\f"));
        }
        throw new IllegalArgumentException("Can not convert value to concrete values: " + value);
    }


}
