package dk.brics.tajs.csi;

import dk.brics.tajs.analysis.ContextArguments;
import dk.brics.tajs.flowgraph.AbstractNode;
import dk.brics.tajs.flowgraph.Function;
import dk.brics.tajs.flowgraph.jsnodes.CallNode;
import dk.brics.tajs.flowgraph.jsnodes.NewObjectNode;
import dk.brics.tajs.flowgraph.jsnodes.ReadVariableNode;
import dk.brics.tajs.lattice.Value;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static dk.brics.tajs.util.Collections.*;

public class ParameterDependencyStrategy implements LiteralContextSensitivityStrategy {
    private final Map<AbstractNode, Set<String>> parameterDependencies = newMap();
    private final Set<AbstractNode> inForIn = newSet();
    // State, assumes DFS AST traversal
    private final Map<Function, Integer> forInDepthCounter = newMap();
    private final LinkedList<AbstractNode> literalStack = new LinkedList<>();

    @Override
    public void registerBeginObjectLiteral(NewObjectNode node) {
        if (FlowGraphUtil.get(node.getBlock().getFunction(), forInDepthCounter) > 0) {
            inForIn.add(node);
        }
        literalStack.push(node);
    }

    @Override
    public void registerEndObjectLiteral(NewObjectNode node) {
        literalStack.pop();
    }

    @Override
    public void registerVariableRead(Function function, ReadVariableNode readVariableNode) {
        if (!literalStack.isEmpty()) {
            if (function.getParameterNames().contains(readVariableNode.getVariableName())) {
                addToMapSet(parameterDependencies, literalStack.peek(), readVariableNode.getVariableName());
            }
        }
    }

    @Override
    public boolean shoudArrayBeSensitive(CallNode node, ContextArguments arguments) {
        return inForIn.contains(node) || shouldBeSensitiveInArguments(node, arguments);
    }

    private boolean shouldBeSensitiveInArguments(AbstractNode node, ContextArguments arguments) {
        if (arguments == null || arguments.isUnknown()) {
            return false;
        }
        List<String> parameterNames = node.getBlock().getFunction().getParameterNames();
        int argumentNumber = 0;
        for (Value value : arguments.getArguments()) {
            boolean isSensitive = value != null;
            boolean hasParameterNameForArgument = parameterNames.size() > argumentNumber;
            if (isSensitive && hasParameterNameForArgument) {
                String parameterName = parameterNames.get(argumentNumber);
                if (parameterDependencies.containsKey(node) && parameterDependencies.get(node).contains(parameterName)) {
                    // the context arguments is sensitive in a parameter (name) that is used in the literal instantiation
//                    System.out.println(node.getSourceLocation().getLineNumber() + ": " + parameterNames + " :: " + arguments.getArguments());
                    return true;
                }
            }
            argumentNumber++;
        }

        return false;
    }

    @Override
    public boolean shoudObjectBeSensitive(NewObjectNode node, ContextArguments arguments) {
        return inForIn.contains(node) || shouldBeSensitiveInArguments(node, arguments);
    }

    @Override
    public void registerBeginForIn(Function function) {
        FlowGraphUtil.begin(function, forInDepthCounter);
    }

    @Override
    public void registerEndForIn(Function function) {
        FlowGraphUtil.end(function, forInDepthCounter);
    }
}
