package dk.brics.tajs.csi;

import dk.brics.tajs.flowgraph.Function;
import dk.brics.tajs.flowgraph.jsnodes.ReadVariableNode;

import java.util.Map;
import java.util.Set;

import static dk.brics.tajs.util.Collections.addToMapSet;
import static dk.brics.tajs.util.Collections.newMap;

public class SyntaxBasedLoopVariableIdentifierManager {

    private static SyntaxBasedLoopVariableIdentifierManager instance;
    private final IdentificationStrategy identifier;

    public SyntaxBasedLoopVariableIdentifierManager(IdentificationStrategy identifier) {
        this.identifier = identifier;
    }

    public static void init(IdentificationStrategy identifier) {
        instance = new SyntaxBasedLoopVariableIdentifierManager(identifier);
    }

    public static void registerBeginLoop(Function function) {
        instance.identifier.registerBeginLoop(function);
    }

    public static void registerEndLoop(Function function) {
        instance.identifier.registerEndLoop(function);
    }

    public static void registerBeginCondition(Function function) {
        instance.identifier.registerBeginCondition(function);
    }

    public static void registerEndCondition(Function function) {
        instance.identifier.registerEndCondition(function);
    }

    public static void registerVariableRead(Function function, ReadVariableNode readVariableNode) {
        instance.identifier.registerVariableRead(function, readVariableNode);
    }

    public static void registerBeginDynamicPropertyRead(Function function) {
        instance.identifier.registerBeginDynamicPropertyRead(function);
    }

    public static void registerEndDynamicPropertyRead(Function function) {
        instance.identifier.registerEndDynamicPropertyRead(function);
    }


    public interface IdentificationStrategy {

        boolean isLoopVariable(Function function, String variable);

        void registerBeginDynamicPropertyRead(Function function);

        void registerBeginLoop(Function function);

        void registerEndDynamicPropertyRead(Function function);

        void registerEndLoop(Function function);

        void registerEndCondition(Function function);

        void registerBeginCondition(Function function);

        void registerVariableRead(Function function, ReadVariableNode readVariableNode);
    }


    public static class ReadInAConditionAndDPAUseInBody implements IdentificationStrategy {
        private final int nestingLimit = 1;
        private final Map<Function, Set<String>> bodyReads;
        private final Map<Function, Set<String>> conditionReads;
        // State, assumes DFS AST traversal
        private final Map<Function, Integer> loopDepthCounter;
        private final Map<Function, Integer> dpaDepthCounter;
        private final Map<Function, Integer> conditionDepthCounter;


        public ReadInAConditionAndDPAUseInBody() {
            loopDepthCounter = newMap();
            dpaDepthCounter = newMap();
            conditionDepthCounter = newMap();
            bodyReads = newMap();
            conditionReads = newMap();
        }

        public boolean isLoopVariable(Function function, String variable) {
            Set<String> conditionReads = this.conditionReads.get(function);
            Set<String> bodyReads = this.bodyReads.get(function);
            return bodyReads != null && conditionReads != null && bodyReads.contains(variable) && conditionReads.contains(variable);
        }

        @Override
        public void registerBeginDynamicPropertyRead(Function function) {
            Map<Function, Integer> counter = dpaDepthCounter;
            FlowGraphUtil.begin(function, counter);
        }


        @Override
        public void registerEndLoop(Function function) {
            FlowGraphUtil.end(function, loopDepthCounter);

        }

        @Override
        public void registerEndCondition(Function function) {
            FlowGraphUtil.end(function, conditionDepthCounter);
        }

        @Override
        public void registerBeginCondition(Function function) {
            FlowGraphUtil.begin(function, conditionDepthCounter);
        }

        @Override
        public void registerVariableRead(Function function, ReadVariableNode readVariableNode) {
            if (loopDepthCounter.containsKey(function) && loopDepthCounter.get(function) > 0) {
                if (conditionDepthCounter.containsKey(function) && conditionDepthCounter.get(function) == 1) { // assumes conditions can not be nested
                    if (loopDepthCounter.get(function) <= nestingLimit) {
                        addToMapSet(conditionReads, function, readVariableNode.getVariableName());
                    }
                } else {
                    if (dpaDepthCounter.containsKey(function) && dpaDepthCounter.get(function) > 0) {
                        addToMapSet(bodyReads, function, readVariableNode.getVariableName());
                    }
                }
            }
        }

        @Override
        public void registerBeginLoop(Function function) {
            FlowGraphUtil.begin(function, loopDepthCounter);
        }

        @Override
        public void registerEndDynamicPropertyRead(Function function) {
            FlowGraphUtil.end(function, dpaDepthCounter);
        }
    }
}
