package dk.brics.tajs.monitoring;

import java.util.Map;
import java.util.Set;

import dk.brics.tajs.Main;
import dk.brics.tajs.analysis.Context;
import dk.brics.tajs.analysis.FunctionCalls;
import dk.brics.tajs.analysis.State;
import dk.brics.tajs.flowgraph.*;
import dk.brics.tajs.flowgraph.jsnodes.IfNode;
import dk.brics.tajs.flowgraph.jsnodes.Node;
import dk.brics.tajs.flowgraph.jsnodes.ReadPropertyNode;
import dk.brics.tajs.flowgraph.jsnodes.ReadVariableNode;
import dk.brics.tajs.flowgraph.jsnodes.WriteVariableNode;
import dk.brics.tajs.lattice.BlockState;
import dk.brics.tajs.lattice.CallEdge;
import dk.brics.tajs.lattice.HostObject;
import dk.brics.tajs.lattice.ObjectLabel;
import dk.brics.tajs.lattice.Str;
import dk.brics.tajs.lattice.Value;
import dk.brics.tajs.monitoring.TypeCollector.VariableSummary;
import dk.brics.tajs.solver.CallGraph;
import dk.brics.tajs.solver.IContext;
import dk.brics.tajs.solver.IMonitoring;
import dk.brics.tajs.solver.Message;
import dk.brics.tajs.solver.Message.Severity;

public interface IAnalysisMonitoring<BlockStateType extends BlockState<BlockStateType,ContextType,CallEdgeType>,
ContextType extends IContext<ContextType>,
CallEdgeType extends CallEdge<BlockStateType>> extends IMonitoring<BlockStateType, ContextType> {

    /**
     * Registers the flowgraph which will be used by the analysis
     */
    void registerFlowgraph(FlowGraph fg);

    void setFinalAnalysisState(State finalState);

    void setFinalCallGraph(CallGraph<State, Context, CallEdge<State>> callGraph);

    /** NB {@link dk.brics.tajs.monitoring.Monitoring#visitNativeFunctionCall(dk.brics.tajs.flowgraph.AbstractNode, dk.brics.tajs.lattice.HostObject, boolean, int, int, int)}  has a similar name.
     * But it is used purely for checking arity... */
    void visitHostFunctionCall(AbstractNode sourceNode, HostObject hostobject, FunctionCalls.CallInfo call, BlockStateType state);

    /**
	 * Checks whether an absent variable is read.
	 * @param n (non-this) read variable operation
	 * @param v the value being read
	 */
	public abstract void visitReadNonThisVariable(ReadVariableNode n, Value v);

	/**
	 * Checks whether the read of 'this' yields the global object.
	 * @param n (this) read variable operation
	 * @param v the value being read
	 * @param state
	 */
	public abstract void visitReadThis(ReadVariableNode n, Value v, BlockStateType state, ObjectLabel global_obj);

	/**
	 * Checks whether the variable read yields null/undefined.
	 * Variables named 'undefined' are ignored.
	 * @param n read variable operation
	 * @param v the value being read
	 */
	public abstract void visitReadVariable(ReadVariableNode n, Value v, BlockStateType state);

	/**
	 * Checks whether the assignment has no effect.
	 * @param n write variable operation
	 * @param v value being written
	 * @param state current abstract state
	 */
	public abstract void visitWriteVariable(WriteVariableNode n, Value v, BlockStateType state);

	/**
	 * Checks whether the branch condition is always true or always false.
	 * @param n if node
	 * @param v the boolean value
	 */
	public abstract void visitIf(IfNode n, Value v);

	/**
	 * Checks whether the 'instanceof' operation may fail.
	 * @param n node performing the operation
	 * @param maybe_v2_non_function set if the second parameter may be a non-function value
	 * @param maybe_v2_function set if the second parameter may be a function value
	 * @param maybe_v2_prototype_primitive set if the prototype property of the second parameter may be a primitive value
	 * @param maybe_v2_prototype_nonprimitive set if the prototype property of the second parameter may be an object value
	 */
	public abstract void visitInstanceof(AbstractNode n,
			boolean maybe_v2_non_function, boolean maybe_v2_function,
			boolean maybe_v2_prototype_primitive, boolean maybe_v2_prototype_nonprimitive);

	/**
	 * Checks whether the 'in' operation may fail.
	 * @param n node performing the operation
	 * @param maybe_v2_object if the second parameter may be an object value
	 * @param maybe_v2_nonobject if the second parameter may be a non-object value
	 */
	public abstract void visitIn(AbstractNode n, boolean maybe_v2_object, boolean maybe_v2_nonobject);

	/**
	 * Checks whether the property access operation may dereference null or undefined.
	 * @param n operation that accesses a property
	 * @param baseval base value for the access
	 */
	public abstract void visitPropertyAccess(Node n, Value baseval);

	/**
	 * Checks whether the property read operation accesses an absent property and whether the operation returns null/undefined.
	 * @param n read property operation
	 * @param objlabels objects being read from
	 * @param propertystr description of the property name
	 * @param maybe if there may be more than one value
	 * @param state current abstract state
	 */
	public abstract void visitReadProperty(ReadPropertyNode n, Set<ObjectLabel> objlabels, Str propertystr, boolean maybe, BlockStateType state);

	/**
	 * Warns about reads from unknown properties;
	 * also registers a read operation on abstract objects.
	 * Properties named 'length' on array objects are ignored.
	 * @param n the node responsible for the read
	 * @param objs the objects being read from
	 * @param propertystr description of the property name
	 * @param check_unknown if set, warn about reads from unknown properties
	 */
	public abstract void visitPropertyRead(Node n, Set<ObjectLabel> objs, Str propertystr, BlockStateType state, boolean check_unknown);

	/**
	 * Warns about writes to unknown properties; 
	 * also registers a write operation on abstract objects.
	 * Properties named 'length' on array objects are ignored.
	 * Writes to the arguments object are also ignored.
	 * @param n the node responsible for the write
	 * @param objs the objects being written to
	 * @param propertystr description of the property name
	 */
	public abstract void visitPropertyWrite(Node n, Set<ObjectLabel> objs, Str propertystr);

	/**
	 * Checks whether the function is invoked both as a constructor (with 'new') and as a function/method (without 'new').
	 * @param f function being called
	 * @param call node responsible for the call
	 * @param constructor if set, the call uses 'new'
	 */
	public abstract void visitUserFunctionCall(Function f, AbstractNode call, boolean constructor);

	/**
	 * Registers a potential call/construct to a non-function value.
	 * @param n node responsible for the call
	 * @param maybe_non_function if set, this call may involve a non-function value
	 * @param maybe_function if set, this call may involve a function value
	 */
	public abstract void visitCall(AbstractNode n, boolean maybe_non_function, boolean maybe_function);

	/**
	 * Registers a call to eval.
	 * @param n node that may call eval
	 * @param v value being eval'ed
	 */
	public abstract void visitEvalCall(AbstractNode n, Value v);

	/**
	 * Registers a write to innerHTML.
	 * @param n node where the write occurs
	 * @param v value being written
	 */
	public abstract void visitInnerHTMLWrite(Node n, Value v);

	/**
	 * Checks the number of parameters for a call to a native function.
	 * @param n node responsible for the call
	 * @param hostobject the native function being called
	 * @param num_actuals_unknown if set, the number of actuals is unknown
	 * @param num_actuals number of actuals (if num_actuals_unknown is not set)
	 * @param min minimum number of parameters expected
	 * @param max maximum number of paramaters expected (-1 for any number)
	 */
	public abstract void visitNativeFunctionCall(AbstractNode n, HostObject hostobject, boolean num_actuals_unknown, int num_actuals, int min, int max);

	/**
	 * Record type information about a var/prop read.
	 */
	public abstract void visitRead(Node n, Value v, BlockStateType state);

	/**
	 * Registers that the given variable is read; also checks for suspicious type mixings.
	 * @param n (non-this) read variable operation 
	 * @param v value being read
	 */
	public abstract void visitVariableAsRead(ReadVariableNode n, Value v, BlockStateType state);

    /**
	 * Produces output showing the parts of the code that is determined unreachable by the analysis.
	 */
	public abstract void generateUnreachableMap();

	/**
	 * Registers the name, location, and value of a variable being read or written.
	 */
	public abstract void visitVariableOrProperty(String var, SourceLocation loc, Value value, IContext context, State state);

	/**
	 * Adds a message for the given node. If not in scan phase, nothing is done. 
	 * Uses Status.INFO.
	 * Uses the message as key (must be a fixed string).
	 */
	public abstract void addMessageInfo(AbstractNode n, Severity severity, String msg);

	/**
	 * Adds a message for the given node. If not in scan phase, nothing is done. 
	 * Uses Status.MAYBE.
	 * Uses the message as key (must be a fixed string).
	 */
	public abstract void addMessage(AbstractNode n, Severity severity, String msg);

	/**
	 * Adds a message for the given node. If not in scan phase, nothing is done. 
	 * Uses Status.MAYBE.
	 * The key must be a fixed string.
	 */
	public abstract void addMessage(AbstractNode n, Severity severity, String key, String msg);

}
