package dk.brics.tajs.monitoring;

import dk.brics.tajs.Main;
import dk.brics.tajs.util.Pair;

import java.util.Map;
import java.util.Stack;

import static dk.brics.tajs.Main.Phase;
import static dk.brics.tajs.util.Collections.newMap;

/**
 * Multi-purpose timer map. Type-wise it is a bastard as it can be used from monitoring as well as anywhere else.
 */
public class TimerMonitoring extends DefaultIAnalysisSideEffectFreeMonitoring {
    private final Map<Timer, Stack<Pair<Long, Long>>> timings = newMap();
    private final Map<Timer, Integer> timerNestings = newMap();

    @Override
    public void beginPhase(Main.Phase phase) {
        registerTimerChange(convertTAJSPhase(phase), true);
    }

    @Override
    public void endPhase(Main.Phase phase) {
        registerTimerChange(convertTAJSPhase(phase), false);
    }

    public void start(Timer timer) {
        registerTimerChange(timer, true);
    }

    public void stop(Timer timer) {
        registerTimerChange(timer, false);
    }

    /**
     * The time TAJS has spent in a specific timer, in nano-seconds.
     */
    public long getElapsedTimeNano(Timer timer) {
        if (!timings.containsKey(timer)) {
            return 0;
        }

        long sum = 0;
        for (Pair<Long, Long> timing : this.timings.get(timer)) {
            sum += timing.getSecond().longValue() - timing.getFirst().longValue();
        }

        return sum;
    }

    /**
     * Register the start/end of a timer
     */

    private void registerTimerChange(Timer timer, boolean start /* false -> end*/) {
        if (!timerNestings.containsKey(timer)) {
            timerNestings.put(timer, 0);
            timings.put(timer, new Stack<Pair<Long, Long>>());
        }
        Integer currentNesting = timerNestings.get(timer);
        if (start) {
            if (currentNesting == 0) { // ignores nested starts of the same timer
                timings.get(timer).push(Pair.<Long, Long>make(System.nanoTime(), null));
            }
            timerNestings.put(timer, currentNesting + 1);
        } else {
            if (currentNesting == 1) { // ignores nested ends of the same timer
                Stack<Pair<Long, Long>> timing = timings.get(timer);
                timing.push(Pair.make(timing.pop().getFirst(), System.nanoTime()));
            }
            timerNestings.put(timer, currentNesting - 1);
        }

    }


    public static enum Timer {PHASE_LOADING_FILES, PHASE_DATAFLOW_ANALYSIS, PHASE_MESSAGES, PHASE_STATISTICS, PHASE_COVERAGE_SUMMARY, PHASE_REACHABLE_VARIABLES_SUMMARY, PHASE_CALL_GRAPH, STRUCTURE_DYNAMIC_OBJECT_LABEL_STATE}

    public static Timer convertTAJSPhase(Phase phase) {
        switch (phase) {
            case LOADING_FILES:
                return Timer.PHASE_LOADING_FILES;
            case DATAFLOW_ANALYSIS:
                return Timer.PHASE_DATAFLOW_ANALYSIS;
            case MESSAGES:
                return Timer.PHASE_MESSAGES;
            case STATISTICS:
                return Timer.PHASE_STATISTICS;
            case COVERAGE_SUMMARY:
                return Timer.PHASE_COVERAGE_SUMMARY;
            case REACHABLE_VARIABLES_SUMMARY:
                return Timer.PHASE_REACHABLE_VARIABLES_SUMMARY;
            case CALL_GRAPH:
                return Timer.PHASE_CALL_GRAPH;
            default:
                throw new UnsupportedOperationException("Unhandled enum: " + phase);
        }
    }
}
