/*
 * Copyright 2009-2013 Aarhus University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package dk.brics.tajs.monitoring;

import java.io.Serializable;
import java.util.LinkedHashMap;
import java.util.Map;

import dk.brics.tajs.flowgraph.SourceLocation;
import dk.brics.tajs.lattice.Value;
import dk.brics.tajs.solver.IContext;


/**
 * Collects abstract values for variable occurrences during the scanning phase of the analysis.
 */
public class TypeCollector { // Used by the TAJS Eclipse plug-in

    /**
     * Tuple of a variable name, source location and a context
     */
    @SuppressWarnings("serial")
    public static class VariableSummary implements Serializable {

        private final String variableName;

        private final SourceLocation sourceLocation;

        private final IContext context;

        public VariableSummary(String variableName, SourceLocation sourceLocation, IContext context) {
            this.variableName = variableName;
            this.sourceLocation = sourceLocation;
            this.context = context;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            VariableSummary other = (VariableSummary) obj;
            if (context == null) {
                if (other.context != null)
                    return false;
            } else if (!context.equals(other.context))
                return false;
            if (sourceLocation == null) {
                if (other.sourceLocation != null)
                    return false;
            } else if (!sourceLocation.equals(other.sourceLocation))
                return false;
            if (variableName == null) {
                if (other.variableName != null)
                    return false;
            } else if (!variableName.equals(other.variableName))
                return false;
            return true;
        }

        public IContext getContext() {
            return context;
        }

        public SourceLocation getSourceLocation() {
            return sourceLocation;
        }

        public String getVariableName() {
            return variableName;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + ((context == null) ? 0 : context.hashCode());
            result = prime * result + ((sourceLocation == null) ? 0 : sourceLocation.hashCode());
            result = prime * result + ((variableName == null) ? 0 : variableName.hashCode());
            return result;
        }

    }

    private Map<VariableSummary, Value> type_info_map = new LinkedHashMap<>();

    /**
     * Records a variable name along with its source location and abstract value.
     *
     * @param variable_name
     *            the name of the variable to record
     * @param source_location
     *            the location in the source code
     * @param value
     *            the abstract value of the variable
     */
    public void record(String variable_name, SourceLocation source_location, Value value, IContext context) { // TODO: do we need to consider unknown or polymorphic values here?
        VariableSummary new_location = new VariableSummary(variable_name == null ? "null" : variable_name, source_location, context);
        Value existing_value = type_info_map.get(new_location);
        if (!value.isPolymorphic()) {
            if (existing_value != null) // TODO: return list of values instead of joining them?
                value = Value.join(existing_value, value); // joining values to accommodate for multiple contexts or imprecise location information
        }
        type_info_map.put(new_location, value);
    }

    /**
     * Returns the collected type information.
     */
    public Map<VariableSummary, Value> getTypeInformation() { // Used by the TAJS Eclipse plug-in
        return type_info_map;
    }

    /**
     * Presents the collected type information using {@code System.out.println} in the format
     * varname: location -> type
     */
    public void presentTypeInformation() { // TODO: present object type information in some better way?
        for (Map.Entry<VariableSummary, Value> entry : type_info_map.entrySet()) {
            final VariableSummary key = entry.getKey();
            System.out.println(key.getVariableName() + ":\t" + key.getSourceLocation() + "\t->\t" + entry.getValue());
        }
    }

    public TypeCollector copy() {
        TypeCollector copy = new TypeCollector();
        copy.type_info_map.putAll(this.type_info_map);
        return copy;
    }

    /**
     * Creates a positive diff of two type collectors. That requires the second typecollector to subsume the first.
     */
    public static TypeCollector makeDiff(TypeCollector first, TypeCollector second) {
        TypeCollector diff = new TypeCollector();
        for (Map.Entry<VariableSummary, Value> entry : second.type_info_map.entrySet()) {
            VariableSummary key = entry.getKey();
            final Value diffValue;
            if (first.type_info_map.containsKey(key)) {
                diffValue = entry.getValue().remove(first.type_info_map.get(key));
            } else {
                diffValue = entry.getValue();
            }
            if(!diffValue.isNone()){
                diff.type_info_map.put(key, diffValue);
            }
        }
        return diff;
    }
}
