package dk.brics.tajs.options;

import com.sun.org.apache.xerces.internal.impl.dv.util.Base64;

import java.io.*;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

/**
 * Utility class for TAJS variants. Git-wise: changes should not be merged upstream to master.
 */
public class ExperimentalOptions implements Serializable {

    private final Set<Option> enabled = new HashSet<>();

    public ExperimentalOptions(Collection<Option> enabled) {
        this.enabled.addAll(enabled);
        this.enabled.add(Option.CONCRETE_SEMANTICS); // TODO ON by default!!
    }

    public ExperimentalOptions(Option... enabled) {
        this(Arrays.asList(enabled));
    }

    public ExperimentalOptions(ExperimentalOptions experimental) {
        this();
        this.enabled.addAll(experimental.enabled);
    }

    public void disable(Option... enabled) {
        this.enabled.removeAll(Arrays.asList(enabled));
    }

    public void enable(Option... enabled) {
        this.enabled.addAll(Arrays.asList(enabled));
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final ExperimentalOptions that = (ExperimentalOptions) o;

        if (!enabled.equals(that.enabled)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        return enabled.hashCode();
    }

    public String getEnabledOptionsString() {
        StringBuilder sb = new StringBuilder();
        for (Option option : enabled) {
            if (sb.length() != 0) {
                sb.append(" ");
            }
            sb.append(option.toString());
        }
        return sb.toString();
    }

    public static ExperimentalOptions intersect(ExperimentalOptions a, ExperimentalOptions b) {
        Set<Option> enabled = new HashSet<>();
        enabled.addAll(a.enabled);
        enabled.retainAll(b.enabled);
        return new ExperimentalOptions(enabled);
    }

    public boolean isEnabled(Option option) {
        return enabled.contains(option);
    }

    public static ExperimentalOptions deserializeFromString(String serialized) {
        byte b[] = Base64.decode(serialized);
        try (ByteArrayInputStream bi = new ByteArrayInputStream(b)) {
            ObjectInputStream si = new ObjectInputStream(bi);
            return (ExperimentalOptions) si.readObject();
        } catch (ClassNotFoundException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public String serializeToString() {
        ByteArrayOutputStream bo = new ByteArrayOutputStream();
        try (ObjectOutputStream so = new ObjectOutputStream(bo)) {
            so.writeObject(this);
            so.flush();
            return Base64.encode(bo.toByteArray());
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public enum Option {
        MANUAL_ASSUMES("manual-assume"), CONTEXT_SENSITIVITY_INFERENCE("call-ctx-infer"),
        SYNTACTIC_LOOP_VARIABLE_IDENTIFICATION("syntax-loop-vars-ident"),
        PROPERTY_CONTEXT_SENSITIVE_OBJECT_LITERALS("prop-ctx-infer"), CONCRETE_SEMANTICS("concrete"), STATECHECKER("statechecker"), STRUCTURAL_ABSTRACTION("structural-abstraction"), STRUCTURAL_ABSTRACTION_MAY("structural-abstraction.may"), STRUCTURAL_ABSTRACTION_RECENCY("structural-abstraction-recency"), STRUCTURAL_ABSTRACTION_NON_ARRAY("structural-abstraction-non-array"), CONSTRUCTOR_ABSTRACTION("constructor-abstraction"), STRUCTURAL_ABSTRACTION_PROPERTY_FILTER("property-filter");
        private final String name;

        private Option(String name) {
            this.name = name;
        }

        @Override
        public String toString() {
            return name;
        }
    }

    public static class ExperimentalOptionsManager {
        private static ExperimentalOptions options = new ExperimentalOptions();

        public static ExperimentalOptions get() {
            return options;
        }

        public static void set(ExperimentalOptions options) {
            ExperimentalOptionsManager.options = options;
        }
    }
}
