package dk.brics.tajs.util;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;

/**
 * Handles writing of analysis outputs to files. Enables most-recent and timestamped versions of the same output file.
 */
public class FileOutputter {
    public ObjectOutputStream makeObjectOutputStream(String name, String extension) {
        try {
            Path file = Files.createTempFile(outputDirectory, name, extension);
            return new ObjectOutputStream(new FileOutputStream(file.toFile()));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public class MultiplexedOutputStream extends OutputStream {
        private final Collection<OutputStream> outputs;

        public MultiplexedOutputStream(Collection<OutputStream> outputs) {
            this.outputs = outputs;
        }

        @Override
        public void write(int b) throws IOException {
            for (OutputStream out : outputs) {
                out.write(b);
            }
        }
    }

    public static final Path DEFAULT_OUT = Paths.get("out");
    private final Path outputDirectory;
    private static final SimpleDateFormat formatter = new SimpleDateFormat("HH:mm");

    /**
     * Output to the default output directory
     */
    public FileOutputter() {
        this.outputDirectory = DEFAULT_OUT;
    }

    /**
     * Convert fileName prefix and suffix to concrete files. The files are located in the output directory.
     * <p/>
     * Creates two files:
     * <ul>
     * <li>name: prefix + suffix</li>
     * <li>name: prefix + [TIMESTAMP] + suffix</li>
     * </ul>
     */
    private Collection<Path> getOutputFiles(String fileNamePrefix, String fileNameSuffix) {
        try {
            Files.createDirectories(outputDirectory);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        Path last = outputDirectory.resolve(String.format("%s%s", fileNamePrefix, fileNameSuffix));
        Path stamped = outputDirectory.resolve(String.format("%s-%s%s", fileNamePrefix, formatter.format(new Date()), fileNameSuffix));
        Collection<Path> files = new HashSet<>();
        files.add(last);
        files.add(stamped);
        return files;
    }

    /**
     * Creates the outputstream to write to for the given fileName prefix and suffix.
     */
    public OutputStream makeOutputStream(String fileNamePrefix, String fileNameSuffix) {
        Collection<Path> outputFiles = getOutputFiles(fileNamePrefix, fileNameSuffix);
        Collection<OutputStream> outputStreams = new HashSet<>();
        for (Path file : outputFiles) {
            try {
                outputStreams.add(new FileOutputStream(file.toFile()));
            } catch (FileNotFoundException e) {
                throw new RuntimeException(e);
            }
        }
        return new MultiplexedOutputStream(outputStreams);
    }

    /**
     * Write output to a file according to fileName prefix and suffix.
     */
    public void output(String output, String fileNamePrefix, String fileNameSuffix) {
        final PrintWriter pw = new PrintWriter(makeOutputStream(fileNamePrefix, fileNameSuffix));
        pw.write(output);
        pw.flush();
    }
}
